from __future__ import annotations

import collections
import collections.abc
import itertools
from typing import TYPE_CHECKING, Any, Callable, DefaultDict, Iterator

if TYPE_CHECKING:
    from PIL import Image  # type: ignore

    from ._base import TkWidget
    from ._images import Icon, Pillow2Tcl
    from ._variables import ControlVariable
    from .fonts import Font
    from .timeouts import Timeout


class count:
    """Simplified itertools.count"""

    def __init__(self, start: int = 0) -> None:
        self._count = start

    def __next__(self) -> int:
        self._count += 1
        return self._count

    def __iter__(self) -> Iterator[int]:
        yield self._count

    def __int__(self) -> int:
        return self._count


counts: DefaultDict[Any, Iterator[int]] = collections.defaultdict(lambda: count())


_commands: dict[str, Callable[..., Any]] = {}
_fonts: dict[str, Font] = {}
_images: dict[str, Icon | Pillow2Tcl] = {}
_pil_images: dict[str, Image.Image] = {}
_text_tags: dict[str, Any] = {}
_timeouts: dict[str, Timeout] = {}
_variables: dict[str, ControlVariable] = {}
_widgets: dict[str, TkWidget] = {}


flatten = itertools.chain.from_iterable


def reversed_dict(dictionary: dict) -> dict:
    return {value: key for key, value in dictionary.items()}


def seq_pairs(sequence):
    return zip(sequence[::2], sequence[1::2])



from __future__ import annotations


class TabStop:
    def __init__(self, position: float, align: str = "left") -> None:
        if align == "decimal":
            align = "numeric"

        self.position, self.align = position, align

    def __repr__(self) -> str:
        return f"TabStop({self.position!r}, {self.align!r})"

    def __to_tcl__(self) -> tuple[float, str]:
        return self.position, self.align



from enum import Enum


class Align(Enum):
    Start = 0
    End = 1
    Stretch = 2


class WindowState(Enum):
    Normal = 0
    Minimized = 1
    Maximized = 2
    Closed = 3
    FullScreen = 4
    Hidden = 5


class Orientation(Enum):
    Horizontal = "horizontal"
    Vertical = "vertical"


class ProgressMode(Enum):
    Determinate = "determinate"
    Indeterminate = "indeterminate"


class Location(Enum):
    ...


class Direction(Enum):
    ...


# ?


class ImagePosition(Enum):
    Bottom = "bottom"
    Default = "none"
    ImageOnly = "image"
    Left = "left"
    Overlap = "center"
    Right = "right"
    TextOnly = "text"
    Top = "top"


class Justify(Enum):
    Center = "center"
    Left = "left"
    Right = "right"


class Anchor(Enum):
    Bottom = "s"
    BottomLeft = "sw"
    BottomRight = "se"
    Center = "center"
    Left = "w"
    Right = "e"
    Top = "n"
    TopLeft = "nw"
    TopRight = "ne"


class Wrap(Enum):
    Letter = "char"
    NoWrap = "none"
    Word = "word"


class CaretStyle(Enum):
    Beam = "0"
    Block = "1"


class InactiveCaretStyle(Enum):
    Hidden = "none"
    Hollow = "hollow"
    Solid = "solid"


class Resizable(Enum):
    Not = (False, False)
    Horizontal = (True, False)
    Vertical = (False, True)
    Both = (True, True)


class BackdropEffect(Enum):
    No = 0
    OpaqueColor = 1
    TransparentColor = 2
    Blur = 3
    Acrylic = 4
    Mica = 5



from __future__ import annotations

from typing import Any


class instanceclassmethod:
    def __init__(self, method) -> None:
        self.method = method

    def __get__(self, obj: Any | None, objtype: type):
        def wrapper(*args, **kwargs):
            if obj is None:
                return self.method(objtype, *args, **kwargs)
            return self.method(obj, *args, **kwargs)

        return wrapper


class classproperty:
    def __init__(self, fget):
        if not isinstance(fget, classmethod):
            fget = classmethod(fget)
        self.fget = fget

    def __get__(self, obj: Any | None, objtype: type):
        return self.fget.__get__(obj, objtype)()



from __future__ import annotations

from typing import Iterable

from tukaan._base import InputControl, TkWidget, WidgetBase, XScrollable
from tukaan._data import Bbox
from tukaan._props import cget, config, fg_color, focusable, text_align, width
from tukaan._tcl import Tcl
from tukaan.colors import Color
from tukaan.exceptions import TclError


class TextBox(WidgetBase, InputControl, XScrollable):
    _tcl_class = "ttk::entry"

    fg_color = fg_color
    focusable = focusable
    text_align = text_align
    width = width

    start = 0
    end = "end"

    def __init__(
        self,
        parent: TkWidget,
        *,
        fg_color: str | Color | None = None,
        focusable: bool | None = None,
        hide_chars: bool = False,
        hide_chars_with: str | None = "•",
        text_align: str | None = None,
        user_edit: bool | None = True,
        value: str | None = None,
        width: int | None = None,
    ) -> None:

        self._prev_show_char = hide_chars_with
        if not hide_chars:
            hide_chars_with = None

        WidgetBase.__init__(
            self,
            parent,
            foreground=fg_color,
            justify=text_align,
            show=hide_chars_with,
            state=None if user_edit else "readonly",
            takefocus=focusable,
            width=width,
        )

        self.bind("<FocusOut>", f"+{self._name} selection clear")

        if value:
            self.set(value)

    def __len__(self):
        return len(self.get())

    def __iter__(self) -> Iterable[str]:
        return iter(self.get())

    def __contains__(self, text: str) -> bool:
        return text in self.get()

    def _repr_details(self) -> str:
        value = self.value
        return f"value='{value if len(value) <= 10 else value[:10] + '...'}'"

    def insert(self, index: int | str | None, string: str) -> None:
        if index is None:
            index = "insert"
        elif isinstance(index, int) and index < 0:
            index = f"end - {index} chars"

        Tcl.call(None, self, "insert", index, string)

    def delete(self, start: int | str, end: int | str = "end") -> None:
        Tcl.call(None, self, "delete", start, end)

    def clear(self) -> None:
        Tcl.call(None, self, "delete", 0, "end")

    def __getitem__(self, indices: slice) -> str:
        return Tcl.call(str, self, "get")[indices]

    def get(self, *indices) -> str:
        if indices:
            return self[slice(*indices)]
        return Tcl.call(str, self, "get")

    def set(self, value: str) -> None:
        Tcl.call(None, self, "delete", 0, "end")
        Tcl.call(None, self, "insert", 0, value)

    value = property(get, set)

    def char_bbox(self, index: int | str) -> Bbox:
        return Bbox(*Tcl.call((int,), self, "bbox", index))

    ### Properties ###

    @property
    def hide_chars_with(self) -> str:
        return cget(self, str, "-show")

    @hide_chars_with.setter
    def hide_chars_with(self, value: str) -> None:
        config(self, show=value)

    @property
    def hide_chars(self) -> bool:
        return cget(self, str, "-show") != ""

    @hide_chars.setter
    def hide_chars(self, value: bool) -> None:
        if value:
            config(self, show=self._prev_show_char)
        else:
            self._prev_show_char = cget(self, str, "-show")
            config(self, show="")

    @property
    def user_edit(self) -> bool:
        return cget(self, str, "-state") != "readonly"

    @user_edit.setter
    def user_edit(self, value: bool) -> None:
        config(self, state="normal" if value else "readonly")

    @property
    def caret_pos(self) -> int:
        return Tcl.call(int, self, "index", "insert")

    @caret_pos.setter
    def caret_pos(self, value: int) -> None:
        Tcl.call(None, self, "icursor", value)

    @property
    def selection(self) -> tuple[int, int] | None:
        try:
            first = Tcl.call(int, self, "index", "sel.first")
            last = Tcl.call(int, self, "index", "sel.last")
        except TclError:
            return None
        else:
            return first, last

    @selection.setter
    def selection(self, value: tuple[int | str, int | str] | None) -> None:
        if value is None:
            Tcl.call(None, self, "selection", "clear")
        elif isinstance(value, (tuple, list)):
            Tcl.call(None, self, "selection", "range", *value)



from __future__ import annotations

import contextlib
from collections.abc import Iterator
from typing import Callable

from PIL import Image

from tukaan._base import Container, TkWidget, WidgetBase
from tukaan._images import Icon, Pillow2Tcl
from tukaan._props import _convert_padding, _convert_padding_back, focusable
from tukaan._tcl import Tcl
from tukaan.enums import ImagePosition
from tukaan.exceptions import TclError

from .frame import Frame


class Tab(Frame):
    def __init__(
        self,
        title: str | None = None,
        *,
        icon: Icon | Image.Image | None = None,
        image_pos: ImagePosition = ImagePosition.Left,
        margin: int | tuple[int, ...] | None = None,
        padding: int | tuple[int, ...] | None = None,
    ):
        Frame.__init__(self, self._widget, padding=padding)

        self._stored_options = {
            "compound": image_pos,
            "image": icon,
            "padding": _convert_padding(margin),
            "text": title,
        }
        self.append()

    def __repr__(self):
        return f"<TabView.Tab in {self.parent}; tcl_name={self._name}>"

    def select(self) -> None:
        if self not in self._widget:
            self.append()

        Tcl.call(None, self._widget, "select", self)

    def append(self) -> None:
        if self in self._widget:
            self.move(-1)
            return None

        Tcl.call(None, self._widget, "add", self, *Tcl.to_tcl_args(**self._stored_options))
        self._widget.tabs.append(self)

    def move(self, new_index: int) -> None:
        self._widget.tabs.remove(self)
        self._widget.tabs.insert(new_index, self)

        if new_index == -1:
            new_index = "end"  # type: ignore

        Tcl.call(None, self._widget, "insert", new_index, self)

    def hide(self) -> None:
        Tcl.call(None, self._widget, "hide", self)

    def unhide(self) -> None:
        Tcl.call(None, self._widget, "add", self)

    def remove(self) -> None:
        with contextlib.suppress(TclError):
            Tcl.call(None, self._widget, "forget", self)
        self._widget.tabs.remove(self)

    @property
    def enabled(self) -> bool:
        return self._get(str, "state") == "normal"

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._set(state="normal" if value else "disabled")

    @property
    def title(self) -> str | None:
        if self in self._widget:
            return Tcl.call(str, self._widget, "tab", self, "-text")
        else:
            return self._stored_options.get("text")

    @title.setter
    def title(self, value: str) -> None:
        if self in self._widget:
            Tcl.call(None, self._widget, "tab", self, "-text", value)
        self._stored_options["text"] = value

    @property
    def icon(self) -> Icon | Image.Image | None:
        if self in self._widget:
            return Tcl.call(Pillow2Tcl, self._widget, "tab", self, "-image")
        else:
            return self._stored_options.get("image")

    @icon.setter
    def icon(self, value: Icon | Image.Image) -> None:
        if self in self._widget:
            Tcl.call(None, self._widget, "tab", self, "-image", value)
        self._stored_options["image"] = value

    @property
    def image_pos(self) -> str | None:
        if self in self._widget:
            return Tcl.call(str, self._widget, "tab", self, "-compound")
        else:
            return self._stored_options.get("compound")

    @image_pos.setter
    def image_pos(self, value: str) -> None:
        if self in self._widget:
            Tcl.call(None, self._widget, "tab", self, "-compound", value)
        self._stored_options["compound"] = value

    @property
    def margin(self) -> tuple[int, int, int, int] | None:
        if self in self._widget:
            result = Tcl.call(str, self._widget, "tab", self, "-padding")
        else:
            result = self._stored_options.get("padding")

        return _convert_padding_back(result)

    @margin.setter
    def margin(self, value: int | tuple[int, ...] | None) -> None:
        converted = _convert_padding(value)

        if self in self._widget:
            Tcl.call(None, self._widget, "tab", self, "-padding", converted)
        self._stored_options["padding"] = converted


class TabView(WidgetBase, Container):
    _tcl_class = "ttk::notebook"

    focusable = focusable

    def __init__(self, parent: TkWidget, *, focusable: bool | None = None) -> None:
        WidgetBase.__init__(self, parent, takefocus=focusable)

        self.Tab = Tab
        setattr(self.Tab, "_widget", self)

        self.tabs = []

    def __len__(self) -> int:
        return len(self.tabs)

    def __iter__(self) -> Iterator[Tab]:
        return iter(self.tabs)

    def __contains__(self, tab: Tab) -> bool:
        return tab in self.tabs

    def __getitem__(self, index: int) -> Tab:
        return self.tabs[index]

    def _repr_details(self) -> str:
        return f"contains {len(self)} tabs"

    def index(self, tab: Tab) -> int:
        return self.tabs.index(tab)

    @property
    def selected(self) -> Tab | None:
        try:
            selected = Tcl.eval(int, f"{self._name} index [{self._name} select]")
        except TclError:
            return None
        else:
            return self.tabs[selected]

    @selected.setter
    def selected(self, tab: Tab) -> None:
        tab.select()

    def on_tab_change(self, func: Callable[[Tab | None], None]) -> Callable[[], None]:
        def wrapper() -> None:
            func(self.selected)

        self.events.bind("<<NotebookTabChanged>>", wrapper)
        return wrapper

    def enable_keyboard_traversal(self) -> None:
        Tcl.call(None, "ttk::notebook::enableTraversal", self)

    def enable_tab_dragging(self) -> None:
        self.events.bind("<Button1-Motion>", self._on_tab_drag, send_event=True)

    def _on_tab_drag(self, event) -> bool | None:
        x, y = event.x, event.y

        if Tcl.call(str, self, "identify", x, y) == "Notebook.tab":
            # I'm using 'Notebook.tab' here, 'coz when the tab has a big image,
            # 'Notebook.tab' glitches only at first tab, 'label' everywhere
            Tcl.eval(
                None,
                f"{self._name} insert [{self._name} index @{x},{y}] [{self._name} select]",
            )
            return False  # like 'break' in tkinter



from __future__ import annotations

from typing import Callable, Iterable

from tukaan._base import TkWidget, WidgetBase
from tukaan._props import cget, config
from tukaan._tcl import Tcl
from tukaan._utils import _commands
from tukaan.colors import Color
from tukaan.exceptions import TclError

from .textbox import TextBox


class ComboBox(TextBox):
    _tcl_class = "ttk::combobox"

    def __init__(
        self,
        parent: TkWidget,
        values: list[str | float] | None = None,
        *,
        current: int | None = None,
        fg_color: str | Color | None = None,
        focusable: bool | None = None,
        hide_chars: bool | None = False,
        hide_chars_with: str | None = "•",
        on_click: Callable | None = None,
        on_select: Callable | None = None,
        text_align: str | None = None,
        user_edit: bool | None = True,
        visible_rows: int | None = None,
        width: int | None = None,
    ) -> None:

        self._prev_show_char = hide_chars_with
        if not hide_chars:
            hide_chars_with = None

        WidgetBase.__init__(
            self,
            parent,
            foreground=fg_color,
            height=visible_rows,
            justify=text_align,
            postcommand=on_click,
            show=hide_chars_with,
            state=None if user_edit else "readonly",
            takefocus=focusable,
            values=values,
            width=width,
        )

        self.bind("<<ComboboxSelected>>", f"+{self._name} selection clear")
        self.bind("<FocusOut>", f"+{self._name} selection clear")

        if on_select:
            self.bind("<<ComboboxSelected>>", on_select)

        self.current = 0 if current is None else current

    def __len__(self) -> int:
        return len(Tcl.call([str], self, "cget", "-values"))

    def __iter__(self) -> Iterable[str]:
        return iter(Tcl.call([str], self, "cget", "-values"))

    def __contains__(self, item: str) -> bool:
        return item in Tcl.call([str], self, "cget", "-values")

    def set(self, value: str) -> None:
        Tcl.call(None, self, "set", value)

    value = property(TextBox.get, set)

    # Properties #

    @property
    def current(self) -> int:
        return self.values.index(self.get())

    @current.setter
    def current(self, index: int) -> None:
        try:
            Tcl.call(None, self, "current", index)
        except TclError:
            raise IndexError("ComboBox index out of range") from None

    @property
    def visible_rows(self) -> int:
        return cget(self, int, "-height")

    @visible_rows.setter
    def visible_rows(self, value: int) -> None:
        config(self, height=value)

    @property
    def values(self) -> list[str]:
        return cget(self, [str], "-values")

    @values.setter
    def values(self, value: list[str | float] | None) -> None:
        if value is None:
            value = []
        config(self, values=value)

    @property
    def on_click(self) -> Callable[..., None] | None:
        return _commands.get(cget(self, str, "-postcommand"))

    @on_click.setter
    def on_click(self, func: Callable[..., None] | None) -> None:
        value = func
        if value is None:
            value = ""
        config(self, postcommand=value)



from __future__ import annotations

from typing import Generator

from tukaan._base import OutputDisplay, TkWidget, WidgetBase
from tukaan._props import cget, config, focusable, link, orientation, value
from tukaan._tcl import Tcl
from tukaan._variables import Float, Integer
from tukaan.enums import Orientation, ProgressMode
from tukaan.timeouts import Timeout


class ProgressBar(WidgetBase, OutputDisplay):
    _tcl_class = "ttk::progressbar"

    focusable = focusable
    orientation = orientation
    value = value
    link = link

    _timeout = None

    def __init__(
        self,
        parent: TkWidget,
        length: int = 100,
        *,
        focusable: bool | None = None,
        link: Integer | Float | None = None,
        mode: ProgressMode | None = None,
        orientation: Orientation | None = None,
        value: int | None = None,
    ) -> None:
        self._max = length

        WidgetBase.__init__(
            self,
            parent,
            max=length,
            mode=mode,
            orient=orientation,
            takefocus=focusable,
            value=value,
            variable=link,
        )

    def _repr_details(self):
        return f"mode={self.mode!r}, length={self.max!r}, value={self.value!r}"

    def step(self, amount: int = 1) -> None:
        Tcl.call(None, self, "step", amount)

    def start_progress(self, steps_per_second: int = 20) -> None:
        self._timeout = Timeout(1 / steps_per_second, self.step)
        self._timeout.repeat()

    def stop_progress(self) -> None:
        if self._timeout is not None:
            self._timeout.cancel()

    def through(self) -> Generator[int, None, None]:
        self.value = 0
        yield 0

        for i in range(1, self._max):
            Tcl.call(None, self._name, "step")
            yield i

    ### Properties ###

    @property
    def length(self) -> int:
        return cget(self, int, "-max")

    @length.setter
    def length(self, value: int) -> None:
        self._max = value
        return config(self, max=value)

    @property
    def mode(self) -> ProgressMode:
        return cget(self, ProgressMode, "-mode")

    @mode.setter
    def mode(self, value: ProgressMode) -> None:
        return config(self, mode=value)



__all__: list = []  # Making star imports impossible. Is it illegal?
__author__ = "rdbende"
__license__ = "MIT license"
__package__ = "tukaan"
__version__ = "0.1.0"


import sys

from ._events import DragObject, KeySeq
from ._images import Icon, IconFactory, Image
from ._info import Clipboard, Pointer, Screen, System
from ._variables import Boolean, Float, Integer, String
from .app import App
from .colors import Color, cmyk, hsl, hsv, rgb
from .fonts import Font, font
from .screen_distance import ScreenDistance, cm, inch, mm
from .time import Time
from .timeouts import Timeout, Timer
from .widgets.button import Button
from .widgets.checkbox import CheckBox
from .widgets.combobox import ComboBox
from .widgets.frame import Frame
from .widgets.label import Label
from .widgets.progressbar import ProgressBar
from .widgets.radiobutton import RadioButton, RadioGroup
from .widgets.scrollbar import ScrollBar
from .widgets.separator import Separator
from .widgets.slider import Slider
from .widgets.spinbox import SpinBox
from .widgets.splitview import SplitView
from .widgets.tabview import TabView
from .widgets.textbox import TextBox
from .widgets.window import Window


class _RequiredVersion:
    def __splitver(self, version_str):
        return tuple(map(int, version_str.split(".")))

    def __eq__(self, version):
        if version != __version__:
            sys.tracebacklimit = 0
            raise RuntimeError(
                f"you have Tukaan version {__version__} installed, which is incompatible with your code that requires Tukaan {version}.\n"
                + f"Use 'pip install tukaan=={version}' to install the required version of Tukaan."
            )

    def __le__(self, version):
        if self.__splitver(__version__) > self.__splitver(version):
            sys.tracebacklimit = 0
            raise RuntimeError(
                f"you have Tukaan version {__version__} installed, which is incompatible with your code that requires Tukaan {version} or older.\n"
                + f"Use 'pip install tukaan=={version}' to install an older version of Tukaan."
            )

    def __ge__(self, version):
        if self.__splitver(__version__) < self.__splitver(version):
            sys.tracebacklimit = 0
            raise RuntimeError(
                f"you have Tukaan version {__version__} installed, which is incompatible with your code that requires Tukaan {version} or newer.\n"
                + "Use 'pip install --upgrade tukaan' to install the newest version of Tukaan."
            )


required_version = _RequiredVersion()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from ._structures import TabStop
from ._tcl import Tcl
from ._utils import _commands, seq_pairs
from ._variables import ControlVariable
from .colors import Color
from .enums import ImagePosition, Justify, Orientation

if TYPE_CHECKING:
    from ._base import TkWidget


def config(widget: TkWidget, **kwargs) -> None:
    Tcl.call(None, widget, "configure", *Tcl.to_tcl_args(**kwargs))


def cget(widget: TkWidget, return_type: Any, option: str) -> Any:
    return Tcl.call(return_type, widget, "cget", option)


def get_image_pos(self) -> ImagePosition:
    return cget(self, ImagePosition, "-compound")


def set_image_pos(self, value: ImagePosition) -> None:
    config(self, compound=value)


image_pos = property(get_image_pos, set_image_pos)


def get_text_align(self) -> Justify:
    return cget(self, Justify, "-justify")


def set_text_align(self, value: Justify) -> None:
    config(self, justify=value)


text_align = property(get_text_align, set_text_align)


def get_fg_color(self) -> Color:
    return cget(self, Color, "-foreground")


def set_fg_color(self, value: Color | str) -> None:
    config(self, foreground=value)


fg_color = property(get_fg_color, set_fg_color)


def get_bg_color(self) -> Color:
    return cget(self, Color, "-foreground")


def set_bg_color(self, value: Color | str) -> None:
    config(self, foreground=value)


bg_color = property(get_bg_color, set_bg_color)


def get_text(self) -> str:
    return cget(self, str, "-text")


def set_text(self, value: str) -> None:
    config(self, text=value)


text = property(get_text, set_text)


def get_width(self) -> int:
    return cget(self, int, "-width")


def set_width(self, value: int) -> None:
    config(self, width=value)


width = property(get_width, set_width)


def get_height(self) -> int:
    return cget(self, int, "-height")


def set_height(self, value: int) -> None:
    config(self, height=value)


height = property(get_height, set_height)


def get_command(self) -> Callable | None:
    return _commands.get(cget(self, str, "-command"))


def set_command(self, func: Callable | None) -> None:
    value = func
    if value is None:
        value = ""
    config(self, command=value)


command = property(get_command, set_command)


def get_orientation(self) -> Orientation:
    return cget(self, Orientation, "-orient")


def set_orientation(self, value: Orientation) -> None:
    return config(self, orient=value)


orientation = property(get_orientation, set_orientation)


def get_value(self) -> int:
    return cget(self, int, "-value")


def set_value(self, value: int) -> None:
    return config(self, value=value)


value = property(get_value, set_value)


def get_link(self) -> ControlVariable:
    return cget(self, ControlVariable, "-variable")


def set_link(self, value: ControlVariable | None) -> None:
    self._variable = value

    if value is None:
        value = ""  # type: ignore
    return config(self, variable=value)


link = property(get_link, set_link)


def get_focusable(self) -> bool:
    return cget(self, bool, "-takefocus")


def set_focusable(self, value: bool) -> None:
    config(self, takefocus=value)


focusable = property(get_focusable, set_focusable)


def get_tab_stops(self) -> list[TabStop]:
    return [TabStop(pos, align) for pos, align in seq_pairs(cget(self, [str], "-tabs"))]


def set_tab_stops(self, tab_stops: TabStop | list[TabStop]) -> None:
    if tab_stops is None:
        return

    if isinstance(tab_stops, TabStop):
        tab_stops = [tab_stops]

    config(self, tabs=[y for x in tab_stops for y in x.__to_tcl__()])


tab_stops = property(get_tab_stops, set_tab_stops)


def _convert_padding(padding: int | tuple[int, ...] | None) -> tuple | tuple[int, ...] | str:
    if padding is None:
        return ()
    elif isinstance(padding, int):
        return (value,) * 4
    else:
        length = len(padding)
        if length == 1:
            return padding * 4
        elif length == 2:
            return (padding[1], padding[0], padding[1], padding[0])
        elif length == 3:
            return (padding[1], padding[0], padding[1], padding[2])
        elif length == 4:
            return (padding[3], padding[0], padding[1], padding[2])
        else:
            return ""


def _convert_padding_back(padding) -> tuple[int, int, int, int]:
    if len(padding) == 1:
        return (padding[0],) * 4
    elif len(padding) == 4:
        return (padding[1], padding[2], padding[3], padding[0])

    return (0,) * 4


def get_padding(self) -> tuple[int, int, int, int]:
    return _convert_padding_back(cget(self, (int,), "-padding"))


def set_padding(self, value: int | tuple[int, ...] | None) -> None:
    config(self, padding=_convert_padding(value))


padding = property(get_padding, set_padding)



from __future__ import annotations

from typing import Callable

from tukaan._base import InputControl, TkWidget, WidgetBase
from tukaan._props import command, focusable, link, text, value, width
from tukaan._tcl import Tcl
from tukaan._variables import ControlVariable, String
from tukaan.enums import Orientation

from .frame import Frame


class RadioButton(WidgetBase, InputControl):
    _tcl_class = "ttk::radiobutton"

    focusable = focusable
    link = link
    on_click = command
    text = text
    value = value
    width = width

    def __init__(
        self,
        parent: TkWidget,
        text: str,
        value: str | float | bool,
        link: ControlVariable,
        *,
        focusable: bool | None = None,
        on_click: Callable[..., None] | None = None,
        width: int | None = None,
    ) -> None:
        self._variable = link

        WidgetBase.__init__(
            self,
            parent,
            command=on_click,
            takefocus=focusable,
            value=value,
            variable=link,
            width=width,
            text=text,
        )

    def invoke(self):
        """Invokes the radiobutton, as if it were clicked"""

        Tcl.call(None, self, "invoke")

    def select(self):
        """Selects the radiobutton"""

        self._variable.set(self.value)

    @property
    def selected(self) -> bool:
        """Returns if the radiobutton is selected or not"""

        return self._variable.get() == self.value


class RadioGroup(Frame, InputControl):
    _items: dict[str, RadioButton]

    def __init__(
        self,
        parent: TkWidget,
        items: dict[str, str],
        *,
        selected: str | None = None,
        orientation: Orientation = Orientation.Vertical,
        padding: int | tuple[int, ...] | None = None,
    ) -> None:
        self._items = {}

        Frame.__init__(self, parent, padding=padding)

        self.link = String(selected or tuple(items.keys())[0])
        self._orient = orientation
        self._set_items(items)

    def _repr_details(self) -> str:
        item_id = self.link.get()
        if item_id not in self._items:
            item_id = None

        return f"number of items={len(self._items)}, selected={item_id!r}"

    def _set_items(self, items: dict[str, str]) -> None:
        for (value, text) in items.items():
            radio = RadioButton(self, text, value, self.link)
            self._items[value] = radio
        self._regrid()

    def _regrid(self):
        is_vert = self._orient is Orientation.Vertical
        for index, radio in enumerate(self._items.values()):
            if is_vert:
                Tcl.call(None, "grid", radio, "-row", index, "-column", 0, "-sticky", "w")
            else:
                Tcl.call(None, "grid", radio, "-row", 0, "-column", index)

    def __getitem__(self, item: str) -> RadioButton | None:
        return self._items.get(item)

    def append(self, value: str, text: str) -> None:
        radio = RadioButton(self, text, value, self.link)
        self._items[value] = radio
        self._regrid()

    def remove(self, item: str) -> None:
        self._items[item].destroy()

    @property
    def value(self) -> str:
        return self.link.get()

    @value.setter
    def value(self, value: str) -> None:
        self.link.set(value)

    @property
    def selected(self) -> RadioButton | None:
        return self._items.get(self.link.get())

    @selected.setter
    def selected(self, value: str) -> None:
        self.link.set(value)

    @property
    def orientation(self) -> Orientation:
        return self._orient

    @orientation.setter
    def orientation(self, value: str) -> None:
        self._orient = value
        self._regrid()



from __future__ import annotations

from tukaan._base import OutputDisplay, TkWidget, WidgetBase
from tukaan._props import orientation
from tukaan.enums import Orientation


class Separator(WidgetBase, OutputDisplay):
    _tcl_class = "ttk::separator"

    orientation = orientation

    def __init__(
        self,
        parent: TkWidget,
        orientation: Orientation | None = None,
    ) -> None:
        WidgetBase.__init__(self, parent, orient=orientation)

    def _repr_details(self):
        return f"orientation={self.orientation!r}"



from __future__ import annotations

from typing import Callable

from PIL import Image  # type: ignore

from tukaan._base import InputControl, TkWidget, WidgetBase
from tukaan._images import Icon
from tukaan._props import command, focusable, image_pos, text, width
from tukaan._tcl import Tcl
from tukaan.enums import ImagePosition


class Button(WidgetBase, InputControl):
    _tcl_class = "ttk::button"

    focusable = focusable
    image_pos = image_pos
    on_click = command
    text = text
    width = width

    def __init__(
        self,
        parent: TkWidget,
        text: str | None = None,
        on_click: Callable | None = None,
        *,
        focusable: bool | None = None,
        image: Image.Image | Icon | None = None,
        image_pos: ImagePosition | None = None,
        width: int | None = None,
    ) -> None:
        WidgetBase.__init__(
            self,
            parent,
            command=on_click,
            compound=image_pos,
            image=image,
            takefocus=focusable,
            text=text,
            width=width,
        )

    def invoke(self) -> None:
        """Invokes the button, as if it were clicked"""

        Tcl.call(None, self, "invoke")



from __future__ import annotations

import collections
import sys
import traceback
from enum import Enum, EnumMeta
from numbers import Real
from pathlib import Path
from typing import Any, Callable

import _tkinter as tk

from ._utils import _commands, counts, flatten, reversed_dict, seq_pairs
from .exceptions import AppError, TclError

_tcl_interp: tk.tkapp | None = None


class Tcl:
    """A Python interface to the Tcl interpreter.

    Attributes:
        interp_address: Stores the address of the Tcl interpreter.
            This attribute is only ised by PIL.
        version: Stores Tcl version info returned by the `info patchlevel` call.
        windowing_system: Stores the windowing system
            name returned by the `tk windowingsystem` call.

    """

    interp_address = None
    version = None
    windowing_system = None

    @classmethod
    def init(cls) -> None:
        global _tcl_interp

        if _tcl_interp is not None:
            raise AppError("can't create multiple Tcl interpreter.")

        _tcl_interp = tk.create(None, sys.argv[0], "Tukaan", False, False, True, False, None)
        _tcl_interp.loadtk()

        cls.interp_address = _tcl_interp.interpaddr()  # PIL needs this
        cls.windowing_system = _tcl_interp.call("tk", "windowingsystem").lower()
        cls.version = _tcl_interp.call("info", "patchlevel")

    @staticmethod
    def get_interp() -> tk.tkapp:
        if _tcl_interp is None:
            Tcl()

        return _tcl_interp

    @staticmethod
    def to_tcl_args(**kwargs) -> tuple:
        result = []

        for key, value in kwargs.items():
            if value is None:
                continue

            if key.endswith("_"):
                key = key.rstrip("_")

            result.extend([f"-{key}", value])

        return tuple(result)

    @staticmethod
    def from_(type_spec: type[Any], value: str | tk.Tcl_obj) -> Any:
        if type_spec is None:
            return None

        if type_spec is str:
            return Tcl.get_string(value)

        if type_spec is bool:
            return Tcl.get_bool(value) if Tcl.from_(str, value) else None

        if type_spec is int or type_spec is float:
            return Tcl.get_number(type_spec, value)

        if hasattr(type_spec, "__from_tcl__"):
            return type_spec.__from_tcl__(value)

        if isinstance(type_spec, (list, tuple, dict)):
            items = Tcl.get_iterable(value)

            if isinstance(type_spec, list):
                return [Tcl.from_(type_spec[0], item) for item in items]

            if isinstance(type_spec, tuple):
                if len(type_spec) != len(items):
                    type_spec *= len(items)
                return tuple(map(Tcl.from_, type_spec, items))

            if isinstance(type_spec, dict):
                result = {}
                for key, value in seq_pairs(items):
                    result[str(key)] = Tcl.from_(type_spec.get(key, str), value)

                return result

        if isinstance(type_spec, EnumMeta):
            return type_spec(value)

        if type_spec is Path:
            return Path(value).resolve()

    @staticmethod
    def to(value: Any) -> str | tuple | tk.Tcl_Obj:
        if isinstance(value, (str, tk.Tcl_Obj)):
            return value

        if value is None:
            return ""

        if isinstance(value, bool):
            return "1" if value else "0"

        if isinstance(value, Real):
            return str(value)

        if hasattr(value, "_name"):
            return value._name

        if hasattr(value, "__to_tcl__"):
            return value.__to_tcl__()

        if isinstance(value, Enum):
            return str(value.value)

        if isinstance(value, collections.abc.Mapping):
            return tuple(map(Tcl.to, flatten(value.items())))

        if callable(value):
            return Tcl.create_cmd(value)

        if isinstance(value, Path):
            return str(value.resolve())

        try:
            return tuple(map(Tcl.to, value))
        except TypeError:
            raise TypeError(
                "can't convert object to Tcl. Please provide a __to_tcl__ method."
            ) from None

    @staticmethod
    def call(return_type: Any, *args) -> Any:
        try:
            result = _tcl_interp.call(*map(Tcl.to, args))
            if return_type is None:
                return
            return Tcl.from_(return_type, result)

        except tk.TclError as e:
            msg = str(e)

            if "application has been destroyed" in msg:
                raise AppError(
                    "can't invoke Tcl callback. Application has been destroyed."
                ) from None

            if not msg.startswith("couldn't read file"):
                raise TclError(msg) from None
            path = msg.split('"')[1]
            sys.tracebacklimit = 0
            raise FileNotFoundError(f"No such file or directory: {path!r}") from None

    @staticmethod
    def eval(return_type: Any, script: str) -> Any:
        try:
            result = _tcl_interp.eval(script)
        except tk.TclError as e:
            raise TclError(str(e)) from None
        else:
            return Tcl.from_(return_type, result)

    @staticmethod
    def create_cmd(func: Callable, *, args=(), kwargs={}) -> str:
        if func in _commands.values():
            return reversed_dict(_commands)[func]

        name = f"tukaan_command_{next(counts['commands'])}"
        _commands[name] = func

        def real_func(*tcl_callback_args):
            try:
                return func(*tcl_callback_args, *args, **kwargs)
            except Exception:
                # remove unnecessary lines:
                # File "/home/.../_utils.py", line 88, in real_func
                # return func(*args)
                tb = traceback.format_exc().split("\n")[3:]
                print("Traceback (most recent call last):\n", "\n".join(tb))

        _tcl_interp.createcommand(name, real_func)
        return name

    @staticmethod
    def delete_cmd(name: str) -> None:
        _tcl_interp.deletecommand(name)
        del _commands[name]

    @staticmethod
    def get_bool(obj) -> bool:
        return _tcl_interp.getboolean(obj)

    @staticmethod
    def get_string(value: str | tk.Tcl_obj) -> str:
        if isinstance(value, str):
            return value

        if isinstance(value, tk.Tcl_Obj):
            return value.string

        return _tcl_interp.call("format", value)

    @staticmethod
    def get_number(type_spec: type[int] | type[float], value: str | tk.Tcl_obj) -> float | None:
        return type_spec(_tcl_interp.getdouble(value)) if Tcl.get_string(value) else None

    @staticmethod
    def get_iterable(value: str | tk.Tcl_obj) -> tuple:
        return _tcl_interp.splitlist(value)

    @staticmethod
    def main_loop() -> None:
        _tcl_interp.mainloop(0)

    @staticmethod
    def quit_interp() -> None:
        global _tcl_interp
        _tcl_interp.quit()
        _tcl_interp = None

    @staticmethod
    def updated(func: Callable) -> Callable:
        def wrapper(self, *args, **kwargs) -> Any:
            _tcl_interp.eval("update idletasks")
            result = func(self, *args, **kwargs)
            _tcl_interp.eval("update idletasks")
            return result

        return wrapper

    @staticmethod
    def update_before(func: Callable) -> Callable:
        def wrapper(self, *args, **kwargs) -> Any:
            _tcl_interp.eval("update idletasks")
            return func(self, *args, **kwargs)

        return wrapper

    @staticmethod
    def update_after(func: Callable) -> Callable:
        def wrapper(self, *args, **kwargs) -> Any:
            result = func(self, *args, **kwargs)
            _tcl_interp.eval("update idletasks")
            return result

        return wrapper



from __future__ import annotations


class Time:
    def __init__(self, *args, **kwargs):
        if args and not kwargs:
            if len(args) == 3:
                h, m, s = args
                if s is None:
                    h, m, s = (0, 0, h) if m is None else (0, h, m)
            elif len(args) == 2:
                h, m, s = 0, *args
            elif len(args) == 1:
                h, m, s = 0, 0, args[0]
        elif kwargs and not args:
            h = kwargs.pop("hours", 0)
            m = kwargs.pop("minutes", 0)
            s = kwargs.pop("seconds", 0)
        else:
            raise ValueError

        if s >= 60:
            plus_m, s = divmod(s, 60)
            m += plus_m
        if m >= 60:
            plus_h, m = divmod(m, 60)
            h += plus_h

        self.hours = int(h)
        self.minutes = int(m)
        self.seconds = s

    def __class_getitem__(cls, key):
        if isinstance(key, slice):
            return Time(key.start, key.stop, key.step)
        elif isinstance(key, (int, float)):
            return Time(seconds=key)

    def __repr__(self):
        time_values = [self.hours, self.minutes, self.seconds]
        if not time_values[0]:
            del time_values[0]

        return f"Time[{':'.join(map(str, time_values))}]"

    @property
    def total_seconds(self):
        return self.hours * 3600 + self.minutes * 60 + self.seconds

    @staticmethod
    def from_secs(seconds: int):
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        return Time(h, m, s)


if __name__ == "__main__":
    assert Time[3].total_seconds == 3
    assert Time[3:3].total_seconds == 183
    assert Time[3:3:3].total_seconds == 10983

    assert Time.from_secs(123456).total_seconds == 123456



from __future__ import annotations

import struct
from collections import namedtuple
from pathlib import Path
from typing import BinaryIO, Iterator

from tukaan._tcl import Tcl
from tukaan._utils import _fonts, counts, seq_pairs
from tukaan.exceptions import FontError, TclError

nameID_order = (
    "copyright",
    "family",
    "subfamily",
    "unique_ID",
    "full_name",
    "version",
    "post_script_name",
    "trademark",
    "manufacturer",
    "designer",
    "description",
    "manufacturer_URL",
    "designer_URL",
    "license",
    "license_URL",
    "reserved",
    "preferred_family",
    "preferred_subfamily",
    "compatible_full_name",
    "sample_text",
    "post_script_find_font_name",
    "WWS_family",
    "WWS_subfamily",
    "light_bg_palette",
    "dark_bg_palette",
    "variations_post_script_name_prefix",
)

preset_fonts = {
    "TkCaptionFont",
    "TkDefaultFont",
    "TkFixedFont",
    "TkHeadingFont",
    "TkIconFont",
    "TkMenuFont",
    "TkSmallCaptionFont",
    "TkTextFont",
    "TkTooltipFont",
}


class FontNameInfo:
    """A class that parses the font file with the help of `struct.unpack`.

    This is a pure Python implementation, though I translated it from Tcl.
    """

    @classmethod
    def get_from_file(cls, file_path: Path) -> list[dict[str, str]]:
        """Gets info from the specified font file.

        Raises:
            FontError: If the magic-tag in the file header can't be recognized.
        """

        fonts_info = []

        with file_path.open("rb") as file:
            magic_tag = file.read(4)  # First 4 byte is the magic tag

            if magic_tag == b"ttcf":
                for offset in cls.read_ttc_header(file):
                    file.seek(offset + 4)  # +4 to skip subfamily magic tag
                    fonts_info.append(cls.read_simple_font_file(file))
            elif magic_tag in {b"\x00\x01\x00\x00", b"OTTO", b"true", b"typ1"}:
                fonts_info.append(cls.read_simple_font_file(file))
            else:
                raise FontError(f"unrecognized magic-tag for OpenType font: 0x{magic_tag.hex()}")

        return fonts_info

    @staticmethod
    def read_ttc_header(file: BinaryIO) -> tuple[int, ...]:
        """Parses the header of a TrueType collection font
        in order to get the offsets of the contained subfamilies.

        Returns:
            A tuple with the offsets counted in bytes.

        """

        *_, num_fonts = struct.unpack(">HHI", file.read(8))
        return struct.unpack(f">{num_fonts}I", file.read(num_fonts * 4))

    @classmethod
    def read_simple_font_file(cls, file: BinaryIO) -> dict[str, str]:
        """Reads a simple font file, or a subfamily of a TTC font.

        Returns:
            The name info of the font family.
        """

        num_tables, *_ = struct.unpack(">4H", file.read(8))
        for _ in range(num_tables):
            table_name, _, start, _ = struct.unpack(">4sL2I", file.read(16))
            if table_name == b"name":  # We don't care about the other tables
                return cls.read_name_table(file, start)

    @staticmethod
    def read_name_table(file: BinaryIO, start: int) -> dict[str, str]:
        """Parses the name table of the font file.

        Returns:
             A dictionary with the field-value pairs.
        """

        file.seek(start)
        _, num_records, str_offset = struct.unpack(">HHH", file.read(6))

        records = [struct.unpack(">6H", file.read(12)) for _ in range(num_records)]
        name_info = {}
        storage_start = start + str_offset

        for platformID, encID, _, nameID, length, offset in records:
            if nameID > 25:
                # Don't care about invalid records
                break

            file.seek(storage_start + offset)
            value, *_ = struct.unpack(f">{length}s", file.read(length))

            if (platformID, encID) == (1, 0):  # 1 == Macintosh
                encoding = "utf-8"
            elif (platformID, encID) == (3, 1):  # 3 == Windows
                encoding = "utf-16-be"  # Microsoft: All string data for platform 3 must be encoded in UTF-16BE.
            else:
                raise FontError(
                    f"unknown platform or encoding specification in font file: {file.name!r}"
                )

            name_info[nameID_order[nameID]] = value.decode(encoding)
            # If the value already exists, just update it with the newly decoded one

        {name_info.setdefault(key) for key in nameID_order}

        return name_info


class Serif:
    """Interface class between the Serif C extension and Python.

    Attributes:
        loaded_fonts: A dictionary that contains the path of the
            loaded files along with the font families contained in them.
    """

    loaded_fonts: dict[Path, list[str]] = {}

    @classmethod
    def load(cls, file_path: Path) -> list[str]:
        """Loads a font file specified by `file_path`.
        The file may contain multiple font families.
        If the font file is already loaded, does nothing.

        Returns:
            A list of the font family names contained in the file.
        """

        if file_path in cls.loaded_fonts:
            return cls.loaded_fonts[file_path]

        Tcl.call(None, "Serif::load_fontfile", file_path)

        info = FontNameInfo.get_from_file(file_path)
        families = []
        for family in info:
            family_name = family["family"]

            if family_name is None:
                raise FontError(f"invalid or missing info in font file: {file_path.resolve()!s}")

            families.append(family_name)

        cls.loaded_fonts[file_path] = family_name

        return families, info

    @classmethod
    def unload(cls, file_path: Path) -> None:
        """
        Unloads a previously loaded font file specified by `file_path`.
        If the file wasn't loaded, does nothing.
        """

        if file_path in cls.loaded_fonts:
            Tcl.call(None, "Serif::unload_fontfile", file_path)

    @classmethod
    def cleanup(cls) -> None:
        """Unloads all loaded fonts."""

        for font in cls.loaded_fonts:
            cls.unload(font)

    @classmethod
    def get_info(cls, file_path: Path) -> list[dict[str, str]]:
        """Gets info from the specified font file. If it was
        previously loaded with Serif.load, returns the cached info.
        """

        if file_path in cls.loaded_fonts:
            return cls.loaded_fonts[file_path]
        return FontNameInfo.get_from_file(file_path)


class FontInfo(namedtuple("FontInfo", nameID_order)):
    """Stores information about a font family"""

    def __iter__(self) -> Iterator[str, str]:
        """Makes it possible to convert the FontInfo to a dictionary.

        Yields:
            The field-value pairs.
        """

        for key in nameID_order:
            yield key, getattr(self, key)


FontMetrics = namedtuple("FontMetrics", ["ascent", "descent", "line_spacing", "fixed"])


class Font:
    """A mutable font object, that can be linked to a Tukaan widget.

    Attributes:
        info: If the font was loaded from a file, this object gives access
            to it's name information
        path: If the font was loaded from a file, stores the filepath
    """

    info: FontInfo | None = None
    path: Path | None = None

    def __init__(
        self,
        family: str | None = None,
        size: int = 10,
        bold: bool = False,
        italic: bool = False,
        underline: bool = False,
        strikethrough: bool = False,
        *,
        file: Path | None = None,
    ) -> None:
        """__init__

        Args:
            family:
                The font family to be used
            size:
                The font size to be used
            bold:
                Bold font ?
            italic:
                Italic font ?
            underline:
                Underlined font ?
            strikethrough:
                Throughstricken font ?
            file:
                A font file to load the family from.
        """
        if isinstance(family, Path):
            file = family

        if file:
            self.path = file if isinstance(file, Path) else Path(file)
            families, info = Serif.load(self.path)

            if not family:
                if len(families) == 1:
                    family = families[0]
                else:
                    raise FontError(
                        f"must specify font family for font collections. Available subfamilies: {families}"
                    )
            elif family not in families:
                raise FontError(f"the family {family} can't be found in this font file")

            self.info = FontInfo(**info[families.index(family)])
        elif not family:
            family = "TkDefaultFont"

        if family in preset_fonts:
            self._name = family
        else:
            self._name = f"tukaan_font_{next(counts['fonts'])}"

        self.config(family, size, bold, italic, underline, strikethrough)

        _fonts[self._name] = self

    def __repr__(self) -> str:
        return f"Font(family={self.family!r}, size={self.size})"

    def config(
        self,
        family: str | None = None,
        size: int = 10,
        bold: bool = False,
        italic: bool = False,
        underline: bool = False,
        strikethrough: bool = False,
    ) -> None:
        """Configures the fonts parameters"""

        args = Tcl.to_tcl_args(
            family=family,
            size=round(size),
            weight="bold" if bold else "normal",
            slant="italic" if italic else "roman",
            underline=underline,
            overstrike=strikethrough,
        )

        try:
            Tcl.call(None, "font", "create", self._name, *args)
        except TclError:
            # Font already exists in Tcl
            Tcl.call(None, "font", "configure", self._name, *args)

    def __to_tcl__(self) -> str:
        return self._name

    @classmethod
    def __from_tcl__(cls, tcl_value: str) -> Font:
        try:
            return _fonts[tcl_value]
        except KeyError:
            flat_values = Tcl.get_iterable(tcl_value)
            types = {
                "family": str,
                "size": int,
                "bold": bool,
                "italic": bool,
                "underline": bool,
                "strikethrough": bool,
            }

            kwargs = {}
            for key, value in seq_pairs(flat_values):
                key = key.lstrip("-")
                if key == "weight":
                    value = value == "bold"
                    key = "bold"
                elif key == "slant":
                    value = value == "italic"
                    key = "italic"
                elif key == "overstrike":
                    key = "strikethrough"
                kwargs[key] = Tcl.from_(types[key], value)

            return cls(**kwargs)

    def _get(self, type_spec: type[int] | type[str] | type[bool], option: str) -> int | str | bool:
        return Tcl.call(type_spec, "font", "actual", self, f"-{option}")

    def _set(self, option: str, value: int | str | bool) -> None:
        Tcl.call(None, "font", "configure", self._name, f"-{option}", value)

    @property
    def family(self) -> str:
        """Gets or sets the current font family."""
        return self._get(str, "family")

    @family.setter
    def family(self, value: str) -> None:
        self._set("family", value)

    @property
    def size(self) -> int:
        """Gets or sets the current font size."""
        return self._get(int, "size")

    @size.setter
    def size(self, value: int):
        self._set("size", value)

    @property
    def bold(self) -> bool:
        """Gets or sets whether the font should be drawn as bold."""
        return self._get(str, "weight") == "bold"

    @bold.setter
    def bold(self, value: bool) -> None:
        self._set("weight", "bold" if value else "normal")

    @property
    def italic(self) -> bool:
        """Gets or sets whether the font should be drawn as italic."""
        return self._get(str, "slant") == "italic"

    @italic.setter
    def italic(self, value: bool) -> None:
        self._set("slant", "italic" if value else "roman")

    @property
    def underline(self) -> bool:
        """Gets or sets whether the font should have underlining."""
        return self._get(bool, "underline")

    @underline.setter
    def underline(self, value: bool) -> None:
        self._set("underline", value)

    @property
    def strikethrough(self) -> bool:
        """Gets or sets whether the font should be striked through."""
        return self._get(bool, "overstrike")

    @strikethrough.setter
    def strikethrough(self, value: bool) -> None:
        self._set("overstrike", value)

    def delete(self) -> None:
        """Deletes the Python as well as the Tcl font object.
        The font won't be usable again, but the previous uses will be preserved.
        """
        Tcl.call(None, "font", "delete", self._name)
        del _fonts[self._name]

    def measure(self, text: str) -> int:
        """Measures, how wide the text would be with the current font family."""
        return Tcl.call(int, "font", "measure", self, text)

    @property
    def metrics(self) -> FontMetrics:
        """Computes the metrics of the current font family."""

        result = Tcl.call(
            {"-ascent": int, "-descent": int, "-linespace": int, "-fixed": bool},
            "font",
            "metrics",
            self,
        )
        return FontMetrics(*tuple(result.values()))


def font(
    family: str | None = None,
    size: int | None = None,
    bold: bool | None = None,
    italic: bool | None = None,
    underline: bool | None = None,
    strikethrough: bool | None = None,
) -> tuple[str, ...]:
    """This function doesn't create a font object, just converts the
    font description to a form understandable by Tcl.

    >>> t = tukaan.TextBox(app, font=tukaan.font(bold=True))
    """

    weight = {None: None, True: "bold", False: "normal"}[bold]
    slant = {None: None, True: "italic", False: "roman"}[italic]

    return Tcl.to_tcl_args(
        family=family,
        size=size,
        weight=weight,
        slant=slant,
        underline=underline,
        overstrike=strikethrough,
    )



from __future__ import annotations

from tukaan._base import Container, TkWidget, WidgetBase
from tukaan._props import padding, set_padding


class Frame(WidgetBase, Container):
    _tcl_class = "ttk::frame"

    padding = padding

    def __init__(self, parent: TkWidget, padding: int | tuple[int, ...] | None = None) -> None:
        WidgetBase.__init__(self, parent)

        set_padding(self, padding)



from __future__ import annotations

from PIL import Image  # type: ignore

from tukaan._base import OutputDisplay, TkWidget, WidgetBase
from tukaan._images import Icon, image
from tukaan._props import bg_color, cget, config, focusable, image_pos, text, text_align
from tukaan.colors import Color
from tukaan.enums import Anchor, ImagePosition, Justify


class Label(WidgetBase, OutputDisplay):
    _tcl_class = "ttk::label"

    bg_color = bg_color
    focusable = focusable
    image = image
    image_pos = image_pos
    text = text
    text_align = text_align

    def __init__(
        self,
        parent: TkWidget,
        text: str | None = None,
        *,
        bg_color: Color | str | None = None,
        content_align: Anchor = Anchor.Center,
        fg_color: Color | str | None = None,
        focusable: bool | None = None,
        image: Image.Image | Icon | None = None,
        image_pos: ImagePosition | None = None,
        max_line_length: int | None = None,
        text_align: Justify = Justify.Left,
    ) -> None:

        WidgetBase.__init__(
            self,
            parent,
            anchor=content_align,
            background=bg_color,
            compound=image_pos,
            foreground=fg_color,
            image=image,
            justify=text_align,
            takefocus=focusable,
            text=text,
            wraplength=max_line_length,
        )

    @property
    def content_align(self) -> Anchor:
        return cget(self, Anchor, "-anchor")

    @content_align.setter
    def content_align(self, value: Anchor) -> None:
        config(self, anchor=value)

    @property
    def max_line_length(self) -> int:
        return cget(self, int, "-wraplength")

    @max_line_length.setter
    def max_line_length(self, value: int) -> None:
        config(self, wraplength=value)



from __future__ import annotations

from typing import TYPE_CHECKING

from ._data import Bbox
from ._layout import Grid
from ._tcl import Tcl
from ._utils import _widgets

if TYPE_CHECKING:
    from ._base import TkWidget


class WidgetMixin:
    _name: str

    def __repr__(self) -> str:
        klass = type(self).__name__
        details = self._repr_details()
        details = f", {details}" if details else ""

        return f"<{klass} widget at {hex(id(self))}: tcl_name={self._name!r} {details}>"

    def _repr_details(self) -> str:
        # overridden in subclasses
        return ""

    @property
    def id(self) -> int:
        return Tcl.call(int, "winfo", "id", self._name)

    def __to_tcl__(self) -> str:
        return self._name

    @classmethod
    def __from_tcl__(cls, tcl_value: str) -> TkWidget:
        return _widgets[tcl_value]


class DnDMixin:
    _lm_path: str

    def set_drag_target(self) -> None:
        Tcl.call(None, "tkdnd::drop_target", "register", self._lm_path, "*")

    def unset_drag_target(self) -> None:
        Tcl.call(None, "tkdnd::drop_target", "unregister", self._lm_path)

    def set_drag_source(self) -> None:
        Tcl.call(None, "tkdnd::drag_source", "register", self._lm_path, "*")

    def unset_drag_source(self) -> None:
        Tcl.call(None, "tkdnd::drag_source", "unregister", self._lm_path)


class GeometryMixin:
    def _abs_x(self) -> int:
        return Tcl.call(int, "winfo", "rootx", self)

    def _abs_y(self) -> int:
        return Tcl.call(int, "winfo", "rooty", self)

    def _rel_x(self) -> int:
        return Tcl.call(int, "winfo", "x", self)

    def _rel_y(self) -> int:
        return Tcl.call(int, "winfo", "y", self)

    def _width(self) -> int:
        return Tcl.call(int, "winfo", "reqwidth", self)

    def _height(self) -> int:
        return Tcl.call(int, "winfo", "reqheight", self)

    @property
    @Tcl.update_before
    def bbox(self) -> tuple:
        return Bbox(self._abs_x(), self._abs_y(), self._width(), self._height())

    rel_x = property(Tcl.update_before(_rel_x))
    rel_y = property(Tcl.update_before(_rel_y))
    abs_x = property(Tcl.update_before(_abs_x))
    abs_y = property(Tcl.update_before(_abs_y))
    width = property(Tcl.update_before(_width))
    height = property(Tcl.update_before(_height))


class VisibilityMixin:
    @property
    @Tcl.update_before
    def visible(self) -> bool:
        return Tcl.call(bool, "winfo", "ismapped", self)

    @visible.setter
    def visible(self, value: bool) -> None:
        if value:
            self.unhide()
        else:
            self.hide()

    def hide(self):
        if self.layout_manager is Grid:
            self.grid._hide()
        else:
            raise NotImplementedError("cannot hide element with this layout manager")

    def unhide(self):
        if self.layout_manager is Grid:
            self.grid._unhide()
        else:
            raise NotImplementedError("cannot hide element with this layout manager")



from __future__ import annotations

from collections import namedtuple

Bbox = namedtuple("Bbox", ["x", "y", "width", "height"])
Position = namedtuple("Position", ["x", "y"])
Size = namedtuple("Size", ["width", "height"])
Version = namedtuple("Version", ["major", "minor", "patchlevel"])



from __future__ import annotations

from tukaan._base import WindowBase, generate_pathname
from tukaan._tcl import Tcl
from tukaan._utils import _widgets
from tukaan.app import App


class Window(WindowBase):
    _tcl_class = "toplevel"
    _keys = {}
    parent: App | Window

    def __init__(
        self,
        parent: App | Window,
        title: str = "Tukaan",
        *,
        width: int = 300,
        height: int = 200,
    ) -> None:
        assert isinstance(parent, WindowBase), "Window parent must be an App or a Window"

        self._wm_path = generate_pathname(self, parent)
        self._name = self._lm_path = f"{self._wm_path}.frame"
        self.parent = parent
        self.parent._children[self._name] = self

        super().__init__()

        Tcl.call(None, self._tcl_class, self._wm_path)
        Tcl.eval(None, f"pack [ttk::frame {self._name}] -expand 1 -fill both")

        self._set_title(title)
        self._set_size((width, height))

        Tcl.call(None, "bind", self, "<Map>", self._generate_state_event)
        Tcl.call(None, "bind", self, "<Unmap>", self._generate_state_event)
        Tcl.call(None, "bind", self, "<Configure>", self._generate_state_event)

    def wait_till_closed(self) -> None:
        Tcl.call(None, "tkwait", "window", self._wm_path)

    def destroy(self) -> None:
        Tcl.call(None, "destroy", self)

        del self.parent._children[self._name]
        del _widgets[self._name]



from __future__ import annotations

import platform
import sys
from fractions import Fraction

import _tkinter as tk
from screeninfo import get_monitors  # type: ignore

from ._behaviour import classproperty
from ._data import Position, Size, Version
from .exceptions import TclError


class System:
    try:
        os = {"linux": "Linux", "win32": "Windows", "darwin": "macOS"}[sys.platform]
    except KeyError:
        os = sys.platform

    py_version = Version(*map(int, platform.python_version_tuple()))
    tk_version = Version(*map(int, tk.TK_VERSION.split(".")[:2]), None)  # type: ignore

    @classproperty
    def tcl_version(self) -> Version:
        from ._tcl import Tcl

        assert Tcl.version is not None
        return Version(*map(int, Tcl.version.split(".")))

    # @classproperty
    # def win_sys(self) -> str:
    #     from ._tcl import Tcl
    #     assert Tcl.windowing_system is not None
    #
    #     return {"win32": "Win32", "x11": "X11", "aqua": "Quartz"}[Tcl.windowing_system]


class Clipboard:
    def __repr__(self) -> str:
        return f"<tukaan.Clipboard; content: {self.paste()}>"

    def clear(self) -> None:
        from ._tcl import Tcl

        Tcl.call(None, "clipboard", "clear")

    def append(self, content: str) -> None:
        from ._tcl import Tcl

        Tcl.call(None, "clipboard", "append", content)

    def paste(self) -> str | None:
        from ._tcl import Tcl

        try:
            return Tcl.call(str, "clipboard", "get")
        except TclError:
            return None

    def copy(self, content: str) -> None:
        from ._tcl import Tcl

        Tcl.call(None, "clipboard", "clear")
        Tcl.call(None, "clipboard", "append", content)


class Pointer:
    def __repr__(self) -> str:
        return f"<tukaan.Pointer; position: {tuple(self.position)}>"

    @classproperty
    def x(self) -> int:
        from ._tcl import Tcl

        return Tcl.call(int, "winfo", "pointerx", ".")

    @classproperty
    def y(self) -> int:
        from ._tcl import Tcl

        return Tcl.call(int, "winfo", "pointery", ".")

    @classproperty
    def position(self) -> tuple[int, int]:
        return Position(self.x, self.y)


common_resolution_standards: dict[tuple[int, int], str] = {
    (320, 200): "CGA",
    (320, 240): "QVGA",
    (640, 480): "VGA",
    (768, 576): "PAL",
    (800, 480): "WVGA",
    (800, 600): "SVGA",
    (854, 480): "FWVGA",
    (1024, 600): "WSVGA",
    (1024, 768): "XGA",
    (1280, 1024): "SXGA",
    (1280, 720): "HD 720",
    (1280, 768): "WXGA",
    (1280, 800): "WXGA",
    (1400, 1050): "SXGA+",
    (1600, 1200): "UXGA",
    (1680, 1050): "WSXGA+",
    (1920, 1080): "HD 1080",
    (1920, 1200): "WUXGA",
    (2048, 1080): "2K",
    (2048, 1536): "QXGA",
    (2560, 1600): "WQXGA",
    (2560, 2048): "QSXGA",
}

common_aspect_ratios: dict[float, str] = {
    3 / 2: "3:2",
    4 / 3: "4:3",
    5 / 3: "5:3",
    5 / 4: "5:4",
    16 / 10: "16:10",
    16 / 9: "16:9",
    17 / 9: "17:9",
}

common_color_depths: dict[int, str] = {
    1: "monochrome",
    15: "high color",
    16: "high color",
    24: "true color",
    30: "deep color",
    36: "deep color",
    48: "deep color",
}


class Screen:
    try:
        for monitor in get_monitors():
            if monitor.is_primary:
                _width = monitor.width or 0
                _height = monitor.height or 0
                _mm_width = monitor.width_mm or 0
                _mm_height = monitor.height_mm or 0
    except Exception:
        # FIXME: Use Tcl winfo command, tho it's influenced by dpi :(
        _width = 0
        _height = 0
        _mm_width = 0
        _mm_height = 0

    @classproperty
    def dpi(self) -> float:
        from ._tcl import Tcl

        return Tcl.call(float, "winfo", "fpixels", ".", "1i")

    @classproperty
    def ppi(self) -> float:
        screen_diagonal_inch = ((self._mm_width / 25.4) ** 2 + (self._mm_height / 25.4) ** 2) ** 0.5
        screen_diagonal_px = (self._width**2 + self._height**2) ** 0.5

        return screen_diagonal_px / screen_diagonal_inch

    @classproperty
    def width(self):
        from .screen_distance import ScreenDistance

        return ScreenDistance(px=self._width)

    @classproperty
    def height(self):
        from .screen_distance import ScreenDistance

        return ScreenDistance(px=self._height)

    @classproperty
    def size(self) -> Size:
        return Size(self._width, self._height)

    @classproperty
    def area(self) -> float:
        return self._width * self._height

    @classproperty
    def diagonal(self):
        from .screen_distance import ScreenDistance

        return ScreenDistance(px=(self._width**2 + self._height**2) ** 0.5)

    @classproperty
    def color_depth(self) -> int:
        from ._tcl import Tcl

        return Tcl.call(int, "winfo", "screendepth", ".")

    @classproperty
    def color_depth_alias(self) -> str:
        try:
            return common_color_depths[self.color_depth]
        except KeyError:
            return ""

    @classproperty
    def aspect_ratio(self) -> str:
        try:
            return common_aspect_ratios[self._width / self._height]
        except KeyError:
            fraction = Fraction(self._width, self._height)  # reduce the ratio
            return f"{fraction.numerator}:{fraction.denominator}"

    @classproperty
    def resolution_standard(self) -> str:
        try:
            return common_resolution_standards[(self._width, self._height)]
        except KeyError:
            return ""
